/*
 * Copyright 2006 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License")
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.jvnet.jaxb2_commons.tools.xjc.plugin.fluent_api;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.xml.sax.ErrorHandler;

import com.sun.codemodel.JBlock;
import com.sun.codemodel.JClass;
import com.sun.codemodel.JCodeModel;
import com.sun.codemodel.JDefinedClass;
import com.sun.codemodel.JExpr;
import com.sun.codemodel.JForEach;
import com.sun.codemodel.JInvocation;
import com.sun.codemodel.JMethod;
import com.sun.codemodel.JMod;
import com.sun.codemodel.JType;
import com.sun.codemodel.JVar;
import com.sun.tools.xjc.Options;
import com.sun.tools.xjc.Plugin;
import com.sun.tools.xjc.outline.ClassOutline;
import com.sun.tools.xjc.outline.Outline;

/**
 * Support a fluent api in addition to the default (JavaBean) setter methods.<br>
 * <p>
 * The initial idea is simply to add a "with*" method to the generated class 
 * for every "set*" method encountered, 
 * with the only functional difference of returning the class instance, instead of void.
 * <p>
 * <strong>Enhancement on 11 June 2006:</strong><br>
 * Provide fluent setter api for Lists, with support of variable arguments.
 * 
 * This enhancement was suggested by Kenny MacLeod <kennym@kizoom.com>,
 * and endorsed by Kohsuke Kawaguchi <Kohsuke.Kawaguchi@sun.com>.
 * Here is quoted from the original request:
 * <p> 
 * <i>By default, XJC represents Lists by generating a getter method, but no setter.  
 * This is impossible to chain with fluent-api.  
 * How about the plugin generates a withXYZ() method for List properties,
 * taking as it's parameters a vararg list.  For example:
 * 
 *<blockquote><pre>
 * // This method is generated by vanilla XJC
 * public List<OtherType> getMyList() {
 *   if (myList == null) {
 *     myList = new ArrayList<OtherType>();
 *   }
 *   return myList;
 * }
 *
 * // This would be generated by fluent-api
 * public MyClass withMyList(OtherType... values) {
 *   for(OtherType value : values) {
 *     getMyList().add(value);
 *   }
 *   return this;
 * }
 *</pre></blockquote></i>
 * 
 * @author Hanson Char
 */
public class XjcFluentApiPlugin extends Plugin
{
    public static final String GETTER_METHOD_PREFIX = "get";
    public static final String SETTER_METHOD_PREFIX = "set";
    public static final String FLUENT_SETTER_METHOD_PREFIX = "with";
    public static final String PARAMETERIZED_LIST_PREFIX = List.class.getName() + "<";
    
    private static final int SETTER_METHOD_PREFIX_LEN = SETTER_METHOD_PREFIX.length();
    private static final int GETTER_METHOD_PREFIX_LEN = GETTER_METHOD_PREFIX.length();
    
    @Override
    public String getOptionName()
    {
        return "Xfluent-api";
    }

    @Override
    public String getUsage()
    {
        return "  -Xfluent-api        :  enable fluent api for generated code";
    }

    @Override
    public boolean run(Outline outline, 
        @SuppressWarnings("unused") Options opt, 
        @SuppressWarnings("unused") ErrorHandler errorHandler)
    {
        final JType voidType = outline.getCodeModel().VOID;
        // Process every pojo class generated by jaxb
        for (ClassOutline classOutline : outline.getClasses()) {
            JDefinedClass implClass = classOutline.implClass;
            Collection<JMethod> setters = new ArrayList<JMethod>();
            Collection<JMethod> listGetters = new ArrayList<JMethod>();
            // Collect the methods we are interested in 
            // but defer the respective fluent methods creation 
            // to avoid ConcurrentModificationException
            for (JMethod jmethod : implClass.methods()) 
            {
            	if (isSetterMethod(jmethod, voidType))
                    setters.add(jmethod);
            	else if (isListGetterMethod(jmethod))
            		listGetters.add(jmethod);
            }
            // Generate a respective fluent method for each setter method
            for (JMethod jmethod : setters)
                createFluentSetterMethod(implClass, jmethod);
            for (JMethod jmethod : listGetters)
            	createFluentListSetterMethod(implClass, jmethod);
        }
        return true;
    }
    
    /** 
     * Adds a fluent api method, which invokes the given setter method, to the given class.
     * This applies to both simple property setter method and 
     * indexed property setter method.
     */
    private void createFluentSetterMethod(JDefinedClass implClass, JMethod setterMethod)
    {
        String name = setterMethod.name();
        // Create a with* method for the respective set* method.
        int mods = JMod.PUBLIC | setterMethod.mods().getValue() & JMod.FINAL;
        JMethod fluentMethod = implClass.method(mods, implClass,  
            FLUENT_SETTER_METHOD_PREFIX + name.substring(SETTER_METHOD_PREFIX_LEN));
        JVar[] jvars = setterMethod.listParams();
        // jvars.length == 1 means simple property setter method.
        // jvars.length == 2 means indexed property setter method.
        assert jvars.length == 1 || jvars.length == 2;
        // with the same parameter(s) as the set* method
        for (JVar jvar : jvars)
            fluentMethod.param(jvar.mods().getValue(), jvar.type(), jvar.name());
        JBlock jblock = fluentMethod.body();
        // The with* method in turn invoke the setter method
        JInvocation jinvocation = jblock.invoke(setterMethod);
        // passing the same list of arguments
        for (JVar jvar : jvars)
            jinvocation.arg(jvar);
        // and return "this"
        jblock._return(JExpr._this());
        return;
    }

    /**
     * Create a fluent setter method for List, with support of variable arguments.
     */
    private void createFluentListSetterMethod(JDefinedClass implClass, JMethod listGetterMethod)
    {
    	String name = listGetterMethod.name();
        // Create a with* method for the respective List<T> get* method.
        int mods = JMod.PUBLIC | listGetterMethod.mods().getValue() & JMod.FINAL;
        JMethod fluentMethod = implClass.method(mods, implClass,  
                FLUENT_SETTER_METHOD_PREFIX + name.substring(GETTER_METHOD_PREFIX_LEN));
        JType returnJType = listGetterMethod.type();
        // As is already checked in isListGetterMethod(JMethod):
        // 1) the return type must be a subtype of JClass; and
        // 2) the number of type parameters must be 1
        JClass returnJClass = JClass.class.cast(returnJType);
        List<JClass> typeParams = returnJClass.getTypeParameters();
        assert typeParams.size() == 1;
        JClass typeParam = typeParams.get(0);
        // Support variable arguments  
        fluentMethod.varParam(typeParam, "values");
        JBlock body = fluentMethod.body();
        JForEach forEach = body.forEach(typeParam, "value", JExpr.ref("values"));
        JBlock forEachBody = forEach.body();
        JInvocation addInvocation = forEachBody.invoke(JExpr.invoke(listGetterMethod), "add");
        addInvocation.arg(JExpr.ref("value"));
        // and return "this"
        body._return(JExpr._this());
        return;
    }
    
    /**
     * Returns true if the given method is a public non-static setter method that follows
     * the JavaBean convention; false otherwise.
     * The setter method can either be a simple property setter method or
     * an indexed property setter method.
     */
    private boolean isSetterMethod(JMethod jmethod, final JType VOID) 
    {
        // Return type of a setter method is expected to be void.
        if (jmethod.type() == VOID) {
            JVar[] jvars = jmethod.listParams();
            
            switch(jvars.length) {
                case 2:
                    // could be an indexed property setter method.
                    // if so, the first argument must be the index (a primitive int).
                    if (!isInt(jvars[0].type()))
                        return false;
                    // drop thru.
                case 1:
                    // or could be a simple property setter method
                    int mods = jmethod.mods().getValue();

                    if ((mods & JMod.STATIC) == 0
                    &&  (mods & JMod.PUBLIC) == 1)
                    {
                        String methodName = jmethod.name();
                        return methodName.length() > SETTER_METHOD_PREFIX_LEN 
                            && methodName.startsWith(SETTER_METHOD_PREFIX);
                    }
                    break;
            }
        }
        return false;
    }
    
    /**
     * Returns true if the given method is a public non-static getter method that returns
     * a List<T>; false otherwise.
     * 
     * @param jmethod given method
     */
    private boolean isListGetterMethod(JMethod jmethod)
    {
        int mods = jmethod.mods().getValue();
        // check if it is a non-static public method
        if ((mods & JMod.STATIC) == 1
        ||  (mods & JMod.PUBLIC) == 0)
        	return false;
        String methodName = jmethod.name();
        // See if the method name looks like a getter method 
        if (methodName.length() <= GETTER_METHOD_PREFIX_LEN 
        || !methodName.startsWith(GETTER_METHOD_PREFIX))
        	return false;
        // A list getter method will have no argument.
    	if (jmethod.listParams().length > 0)
    		return false;
    	// See if the return type of the method
    	// is a List<T>
   		JType jtype = jmethod.type();
    		
    	if (jtype instanceof JClass)
    	{
    		JClass jclass = JClass.class.cast(jtype);
    		List<JClass> typeParams = jclass.getTypeParameters();
    		
    		if (typeParams.size() != 1)
    			return false;
    		return jclass.fullName().startsWith(PARAMETERIZED_LIST_PREFIX);
    	}
    	return false;
    }
    
    /** Returns true if the given type is a primitive int; false otherwise. */
    private boolean isInt(JType type) 
    {
        JCodeModel codeModel = type.owner();
        return type.isPrimitive() 
            && codeModel.INT.equals(
                JType.parse(codeModel, type.name()));
    }
}
